==> FIAT-L-0.1.1.tar.gz: FIAT-L-0.1.1/FIAT/polynomial.py <==
# Written by Robert C. Kirby
# Copyright 2005 by The University of Chicago
# Distributed under the LGPL license
# This work is partially supported by the US Department of Energy
# under award number DE-FG02-04ER25650

# last modified 2 Feb 2005

import expansions, shapes, Numeric, curry, points, LinearAlgebra
from FirstDerivatives import DerivVar as DV

# this can be stripped out in later versions to
# allow for efficient evaluation of the orthogonal
# functions (e.g. using P_{i} to get P_{i+1})
class PolynomialBase( object ):
    def __init__( self , shape , n ):
        self.bs = expansions.make_expansion( shape , n )
        d = shapes.dimension( shape )
        if n == 0:
            dmats = [ Numeric.array( [ [ 0.0 ] ] , "d" ) ] * d
        else:
            dtildes = [ Numeric.zeros( (len(self.bs),len(self.bs)) , "d" ) \
                      for i in range(0,d) ]

            pts = points.make_points( shape , d , 0 , n + d + 1 )

            v = Numeric.array( [ [ b( x ) for b in self.bs ] for x in pts ] )

            for i in range(0,len(pts)):
                for j in range(0,len(self.bs)):
                    for k in range(0,d):
                        x = point_to_partial( pts[i] , k )
                        foo = self.bs[j]( x )
                        if type( foo ) == type( 1. ):
                            dtildes[k][i,j] = 0.0
                        else:
                            dtildes[k][i,j] = foo[1][0]

            vinv = LinearAlgebra.inverse( v )
            
            self.dmats = [ Numeric.dot( vinv , dtilde ) for dtilde in dtildes ]
        return

    def eval_all( self , x ):
        return Numeric.array( [ b( x ) for b in self.bs ] )

    def tabulate( self , xs ):
        """xs is an iteratable object of points (list, tuple, set).
        returns a list of dictionaries where the ith item is the
        dictionary of the ith basis function values at each x in xs"""
        vals = [{} for b in self.bs]
        for x in xs:
            v = self.eval_all( x ) # array of basis values at x
            for i in range(0,len(self.bs)):
                vals[i][x] = v[i]
        return vals

    def __len__( self ):
        return len( self.bs )
    def __getitem__( self , i ):
        return self.bs[i]

    def deriv_all( self , i ):
        pass

# polynomials, possibly with array-valued coefficients
class Polynomial( object ):
    def __init__( self , base , dof ):
        self.base = base
        self.dof = Numeric.asarray( dof )
        return
    def __call__( self , x ):
        bvals = self.base.eval_all( x )
        return reduce( lambda a,b: a + b , \
                       [ v * c for (v,c) in zip(bvals,self.dof) ] )
    def __getitem__( self , i ):
	if type( self.dof[0] ) == type( 1.0 ):  # scalar-valued
	    if i != 0:
		raise RuntimeError, "Can't get that component."
	    else:
		return self
	else:  # array-valued, extract components of coeffs
	    return Polynomial( self.base , \
			       Numeric.array( [ d[i] \
						for d in self.dof ] ) ) 
    def deriv( self , i ):
	# logic -- if i is a scalar (integer), apply the derivative
	# matrix for direction i to each component
	# if i is a tuple of the form ( ii , index ), take the ith
	# directional derivative of the index^th component
	if type( i ) == type( 1 ):
            D = self.base.dmats[i]
            newcoeff = \
                     [ sum( [ dj * cj for (dj,cj) in zip(d,self.dof) ] ) \
                        for d in D ] 
            return Polynomial( self.base , \
                               Numeric.array( newcoeff ) )
                   
	else:
	    (ii,index) = i         # direction
	    return self[ ii ].deriv( index )

class PolynomialSet( object ):
    """A set of polynomials expressed as linear combinations
    of the set base.  Coeffs is an array."""
    def __init__( self , base , coeffs ):
        """base is either a PolynomialBase instance or else
        a PolynomialSet instance."""
        if isinstance( base , PolynomialBase ):
            self.base = base
            self.coeffs = coeffs
        elif isinstance( base , PolynomialSet ):
            # This might be wrong!!!
            self.base = base.base
            self.coeffs = Numeric.dot( coeffs , base.coeffs )
        else:
            raise RuntimeError, "Illegal base input to PolynomialSet."

        self.dim = len(self.coeffs)

    def eval_all( self , x ):
        bvals = self.base.eval_all( x )
        return Numeric.asarray( [ reduce( lambda a,b: a + b , \
                                          [ v * c for (v,c) in zip(bvals,d) ] ) \
                                  for d in self.coeffs ] )

    # applies the i derivative to each entry of the basis.
    # this should be a matrix-vector product, no?
    def deriv_all( self , i ):
        pass

    def tabulate( self , xs ):
        """xs is an iteratable object of points (list, tuple, set).
        returns a list of dictionaries where the ith item is the
        dictionary of the ith basis function values at each x in xs"""
        # really, I should redo this with a tabulate followed by a
        # matrix-matrix multiplication
        vals = [{} for b in range(0,self.dim)]
        for x in xs:
            v = self.eval_all( x ) # array of basis values at x
            for i in range(0,self.dim):
                vals[i][x] = v[i]
        return vals

    def __len__( self ):
        return len( self.coeffs )

    def __getitem__( self , i ):
        return Polynomial( self.base , self.coeffs[i] )

class OrthogonalPolynomialSet( PolynomialSet ):
    def __init__( self , shape , degree ):
	b = PolynomialBase( shape , degree )
	PolynomialSet.__init__( self , \
				b , \
				Numeric.identity( len( b.bs ) ,"d") )
	return

class FiniteElement( PolynomialSet ):
    def __init__( self , dualbasis , U ):
        self.dualbasis = dualbasis
        self.U = U
        vinv = LinearAlgebra.inverse( outer_product( dualbasis.ls , \
                                                    U ) )
        PolynomialSet.__init__( self , self.U , Numeric.transpose( vinv ) )


def outer_product( flist , plist ):
    """FList is a functional.FunctionalList object.
    PList is a polynomial.PolynomialList object.
    Computes the outer product -- A[i,j] where i is
    the index into FList and j is the index into
    PList.  Used in building vandermonde matrices
    for finite element spaces and for building the constraint
    matrix in taking null spaces"""
    return Numeric.asarray( [ [ f(p) for p in plist ] \
                              for f in flist ] )

def outer_product_expert( flist , plist ):
    """Replacement for outer_product that tabulates all of
    the items in the set of polynomials at the support of the
    functionals and then evaluates via the resulting dictionary."""
    pdicts = plist.tabulate( flist.support )
    return Numeric.asarray( [ [ f.expert_call( pdict ) \
                                for pdict in pdicts ] \
                              for f in flist ] )

def point_to_partial( x , i ):
    tmp = list( x )
    tmp[i] = DV( tmp[i] )
    return tuple( tmp )

def deriv( i , p ):
    return p.deriv( i )

def D( i ):
    return curry.curry( deriv , i )

==> FIAT-L-0.1.2.tar.gz: FIAT-L-0.1.2/FIAT/polynomial.py <==
# Written by Robert C. Kirby
# Copyright 2005 by The University of Chicago
# Distributed under the LGPL license
# This work is partially supported by the US Department of Energy
# under award number DE-FG02-04ER25650

# last modified 17 Feb 2005

import expansions, shapes, Numeric, curry, points, LinearAlgebra

# this can be stripped out in later versions to
# allow for efficient evaluation of the orthogonal
# functions (e.g. using P_{i} to get P_{i+1})
class PolynomialBase( object ):
    def __init__( self , shape , n ):
        self.bs = expansions.make_expansion( shape , n )
	self.shape,self.n = shape,n
        d = shapes.dimension( shape )
        if n == 0:
            dmats = [ Numeric.array( [ [ 0.0 ] ] , "d" ) ] * d
        else:
            dtildes = [ Numeric.zeros( (len(self.bs),len(self.bs)) , "d" ) \
                      for i in range(0,d) ]
            pts = points.make_points( shape , d , 0 , n + d + 1 )
	    v = Numeric.transpose( expansions.tabulators[shape](n,pts) )
            vinv = LinearAlgebra.inverse( v )
	    dtildes = [Numeric.transpose(a) \
		       for a in expansions.deriv_tabulators[shape](n,pts)]
            self.dmats = [ Numeric.dot( vinv , dtilde ) for dtilde in dtildes ]
        return

    def eval_all( self , x ):
        return self.tabulate( Numeric.array([x]))[:,0]

    def tabulate( self , xs ):
	"""xs is an iterable object of points.
	returns an array -- one row for each bf, columns are that
	bf at the points."""
	return expansions.tabulators[self.shape](self.n,Numeric.array(xs))

    def __len__( self ):
        return len( self.bs )
    def __getitem__( self , i ):
        return self.bs[i]


# polynomials, possibly with array-valued coefficients
class Polynomial( object ):
    def __init__( self , base , dof ):
        self.base = base
        self.dof = Numeric.asarray( dof )
        return
    def __call__( self , x ):
        bvals = self.base.eval_all( x )
	return Numeric.dot( self.dof , self.base.eval_all( x ) )
    def __getitem__( self , i ):
	if type( self.dof[0] ) == type( 1.0 ):  # scalar-valued
	    if i != 0:
		raise RuntimeError, "Can't get that component."
	    else:
		return self
	else:  # array-valued, extract components of coeffs
	    return Polynomial( self.base , \
			       Numeric.array( [ d[i] \
						for d in self.dof ] ) ) 
    def deriv( self , i ):
	# logic -- if i is a scalar (integer), apply the derivative
	# matrix for direction i to each component
	# if i is a tuple of the form ( ii , index ), take the ith
	# directional derivative of the index^th component
	if type( i ) == type( 1 ):
            D = self.base.dmats[i]
            newcoeff = \
                     [ sum( [ dj * cj for (dj,cj) in zip(d,self.dof) ] ) \
                        for d in D ] 
            return Polynomial( self.base , \
                               Numeric.array( newcoeff ) )
                   
	else:
	    (ii,index) = i         # direction
	    return self[ ii ].deriv( index )

class PolynomialSet( object ):
    """A set of polynomials expressed as linear combinations
    of the set base.  Coeffs is an array where each row is the
    expansion coefficients of each bf for a particular polynomial."""
    def __init__( self , base , coeffs ):
        """base is either a PolynomialBase instance or else
        a PolynomialSet instance."""
        if isinstance( base , PolynomialBase ):
            self.base = base
            self.coeffs = coeffs
        elif isinstance( base , PolynomialSet ):
            # This might be wrong!!!
            self.base = base.base
            self.coeffs = Numeric.dot( coeffs , base.coeffs )
        else:
            raise RuntimeError, "Illegal base input to PolynomialSet."

        self.dim = len(self.coeffs)

    def eval_all( self , x ):
	return Numeric.dot( self.coeffs, self.base.eval_all( x ) )

    # applies the i derivative to each entry of the basis.
    # this should be a matrix-vector product, no?
    def deriv_all( self , i ):
	return PolynomialSet( self.base ,
			      Numeric.array( [ deriv(i,p).dof \
					       for p in self ] ) )

    def tabulate( self , xs ):
        # really, I should redo this with a tabulate followed by a
        # matrix-matrix multiplication
	return Numeric.dot( self.coeffs ,
			    self.base.tabulate( xs ) )


    def __len__( self ):
        return len( self.coeffs )

    def __getitem__( self , i ):
        return Polynomial( self.base , self.coeffs[i] )

class OrthogonalPolynomialSet( PolynomialSet ):
    def __init__( self , shape , degree ):
	b = PolynomialBase( shape , degree )
	PolynomialSet.__init__( self , \
				b , \
				Numeric.identity( len( b.bs ) ,"d") )
	return

class FiniteElement( PolynomialSet ):
    def __init__( self , dualbasis , U ):
        self.dualbasis = dualbasis
        self.U = U
	vinv = LinearAlgebra.inverse( outer_product( dualbasis, U ) )
        PolynomialSet.__init__( self , self.U , Numeric.transpose( vinv ) )


def outer_product( flist , plist ):
    # flist is an iterable object of functionals
    # plist is an iterable object of polynomials
    # This can be expressed as a matrix-matrix product
    # F P where the rows of F are the vectors associated
    # with each functional and the columns of P
    # are the vectors of expansion coefficients
    return Numeric.dot( flist.mat , \
			Numeric.transpose( plist.coeffs ) )


def deriv( i , p ):
    return p.deriv( i )

def D( i ):
    return curry.curry( deriv , i )
